# Glob

Filename matching and globbing with shell patterns.

`fnmatch(file_name, pattern)` matches according to the local convention.
`fnmatchcase(file_name, pattern)` always takes case into account. The
functions operate by translating the pattern into a regular expression.

The function translate(PATTERN) returns a regular expression corresponding to
PATTERN.

Patterns are Unix shell style:
    *       | matches multiple characters within a directory
    **      | matches multiple characters across directories
    ?       | matches any single character
    [seq]   | matches any character in seq
    [!seq]  | matches any char not in seq


```pony
primitive val Glob
```

## Constructors

### create

```pony
new val create()
: Glob val^
```

#### Returns

* [Glob](glob-Glob) val^

---

## Public Functions

### fnmatch

Tests whether `name` matches `pattern`.

An initial period in `name` is not special.

Both `name` and `pattern` are first case-normalized if the operating system
requires it. If you don't want this, use `fnmatchcase`.


```pony
fun box fnmatch(
  name: String val,
  pattern: String val)
: Bool val
```
#### Parameters

*   name: [String](builtin-String) val
*   pattern: [String](builtin-String) val

#### Returns

* [Bool](builtin-Bool) val

---

### fnmatchcase

Tests whether `name` matches `pattern`, including case.

```pony
fun box fnmatchcase(
  name: String val,
  pattern: String val)
: Bool val
```
#### Parameters

*   name: [String](builtin-String) val
*   pattern: [String](builtin-String) val

#### Returns

* [Bool](builtin-Bool) val

---

### filter

Returns `name` and the matching subgroups for `names` that match `pattern`.

All strings are first case-normalized if the operating system requires it.


```pony
fun box filter(
  names: Array[String val] ref,
  pattern: String val)
: Array[(String val , Array[String val] ref)] val
```
#### Parameters

*   names: [Array](builtin-Array)\[[String](builtin-String) val\] ref
*   pattern: [String](builtin-String) val

#### Returns

* [Array](builtin-Array)\[([String](builtin-String) val , [Array](builtin-Array)\[[String](builtin-String) val\] ref)\] val

---

### translate

Translates a shell `pattern` to a regular expression.
There is no way to quote meta-characters.


```pony
fun box translate(
  pat: String val)
: String ref^
```
#### Parameters

*   pat: [String](builtin-String) val

#### Returns

* [String](builtin-String) ref^

---

### glob

Returns an Array[FilePath] for each path below `root_path` that matches
`pattern`.

The pattern may contain shell-style wildcards. See the type documentation
on `Glob` for details.


```pony
fun box glob(
  root_path: FilePath val,
  pattern: String val)
: Array[FilePath val] ref
```
#### Parameters

*   root_path: [FilePath](files-FilePath) val
*   pattern: [String](builtin-String) val

#### Returns

* [Array](builtin-Array)\[[FilePath](files-FilePath) val\] ref

---

### iglob

Calls `GlobHandler.apply` for each path below `root` that matches
`pattern`.

The pattern may contain shell-style wildcards. See the type documentation
on `Glob` for details.


```pony
fun box iglob(
  root: FilePath val,
  pattern: String val,
  glob_handler: GlobHandler ref)
: None val
```
#### Parameters

*   root: [FilePath](files-FilePath) val
*   pattern: [String](builtin-String) val
*   glob_handler: [GlobHandler](glob-GlobHandler) ref

#### Returns

* [None](builtin-None) val

---

### eq

```pony
fun box eq(
  that: Glob val)
: Bool val
```
#### Parameters

*   that: [Glob](glob-Glob) val

#### Returns

* [Bool](builtin-Bool) val

---

### ne

```pony
fun box ne(
  that: Glob val)
: Bool val
```
#### Parameters

*   that: [Glob](glob-Glob) val

#### Returns

* [Bool](builtin-Bool) val

---

## Private Functions

### _apply_glob_to_walk

```pony
fun box _apply_glob_to_walk(
  pattern: String val,
  compiled_pattern: Regex ref,
  root: FilePath val,
  glob_handler: GlobHandler ref,
  dir: FilePath val,
  entries: Array[String val] ref)
: None val
```
#### Parameters

*   pattern: [String](builtin-String) val
*   compiled_pattern: [Regex](regex-Regex) ref
*   root: [FilePath](files-FilePath) val
*   glob_handler: [GlobHandler](glob-GlobHandler) ref
*   dir: [FilePath](files-FilePath) val
*   entries: [Array](builtin-Array)\[[String](builtin-String) val\] ref

#### Returns

* [None](builtin-None) val

---

